#include <cstring>
#include <cmath>

#include "colors.hh"
#include "strfun.hh"

struct ColorInfo { const char*name; unsigned color; };

static bool FindColorInfo(const ColorInfo&a, const ColorInfo& b)
{
    return std::strcmp(a.name, b.name) < 0;
}

unsigned TranslateColorNameToRGB(const std::string& name)
{
    #define RGB(r,g,b) (((r)<<16) + ((g)<<8) + ((b)))
    std::string nameu = UCase(name);
    if(nameu.size() == 4 && nameu[0] == '#')
    {
        unsigned val = strtol(nameu.c_str()+1, 0, 16);
        unsigned r = (val&0xF00)>>8;  r*=0x11;
        unsigned g = (val&0xF0)>>4;   g*=0x11;
        unsigned b = (val&0xF)>>0;    b*=0x11;
        return RGB(r,g,b);
    }
    if(nameu.size()==7 && nameu[0] == '#')
    {
        return std::strtol(nameu.c_str()+1, 0, 16);
    }
    static const ColorInfo colors[] =
    {
#include "colors.inc"
    };
    
    ColorInfo dummy; dummy.name = nameu.c_str();
    const ColorInfo* endcolors = colors+ (sizeof(colors)/sizeof(*colors));
    
    const ColorInfo* color = std::lower_bound( colors, endcolors, dummy, FindColorInfo);
    if(color != endcolors)
        return color->color;
    return 0xC0C0C0;
    #undef RGB
}

double RgbDiff(unsigned rgb1, unsigned rgb2)
{
    return std::pow( (int)((rgb1>>16)&0xFF) - (int)((rgb2>>16)&0xFF), 2.0)
         + std::pow( (int)((rgb1>> 8)&0xFF) - (int)((rgb2>> 8)&0xFF), 2.0)
         + std::pow( (int)((rgb1>> 0)&0xFF) - (int)((rgb2>> 0)&0xFF), 2.0);
}

unsigned CreateXtermColor(const unsigned rgb)
{
    static const unsigned char sixtable[6] = {0x00,0x5F,0x87,0xAF,0xD7,0xFF};
    #define graytable(n) (8 + 10*(n))  /* n=0..23 */
    unsigned r = (rgb>>16)&0xFF, g = (rgb>>8)&0xFF, b = (rgb>>0)&0xFF;
    
    // Try six-algorithm
    unsigned r6 = r*6/256, g6=g*6/256, b6=b*6/256;
    unsigned six_result = (sixtable[r6]<<16) + (sixtable[g6]<<8) + (sixtable[b6]);
    
    // Try grayscale-algorithm
    unsigned gray_value  = (r<8) ? 0 : (r>0xEE ? 23 : (r-8)/10);
    unsigned gray_result = graytable(gray_value);
    
    if(RgbDiff(gray_result,rgb) < RgbDiff(six_result,rgb))
    {
        return 232 + gray_value;
    }
    return 16 + b6 + g6*6 + r6*6*6;
}
